% dal - dual augmented Lagrangian method for sparse learaning/reconstruction
%
% Overview:
%  Solves the following optimization problem
%   xx = argmin f(x) + lambda*c(x)
%  where f is a user specified (convex, smooth) loss function and c
%  is a measure of sparsity (currently L1 or grouped L1)
%
% Syntax:
%  [xx, status] = dal(prob, xx0, A, AT, lambda, <opt>)
%
% Inputs:
%  prob   : structure that contains the following fields:
%   .obj      : DAL objective function
%   .floss    : structure with three fields (p: primal loss, d: dual loss, args: arguments to the loss functions)
%   .freg     : function handle to the regularizer
%   .mm       : number of samples (scalar)
%   .nn       : number of unknown variables (scalar)
%   .stopcond : function handle for the stopping condition
%   .hessMult : function handle to the Hessian product function (H*x)
%   .softth : function handle to the "soft threshold" function
%  xx0    : initial solution ([nn,1] or [ns,nc] with ns*nc=nn)
%  A      : function handle to the function A*x.
%  AT     : function handle to the function A'*y
%  lambda : regularization constant (scalar)
%  <opt>  : list of 'fieldname1', value1, 'filedname2', value2, ...
%   aa        : initial Lagrangian multiplier [mm,1] (default zero(mm,1))
%   tol       : tolerance (default 1e-3)
%   maxiter   : maximum number of outer iterations (default 100)
%   eta       : initial barrier parameter (default 1)
%   eps       : initial internal tolerance parameter (default 1e-4)
%   eta_multp : multiplying factor for eta (default 2)
%   eps_multp : multiplying factor for eps (default 0.5)
%   solver    : internal solver. Can be either:
%               'nt'   : Newton method with cholesky factorization (default)
%               'ntsv' : Newton method saves memory (slightly slower)
%               'cg'   : Newton method with PCG
%               'qn'   : Quasi-Newton method
%   display   : display level (0: none, 1: only the last, 2: every
%               outer iteration, (default) 3: every inner iteration)
%   iter      : output the value of xx at each iteration 
%               (boolean, default 0)
% Outputs:
%  xx     : the final solution
%  status : various status values
%
% Reference:
% "Dual Augmented Lagrangian Method for Efficient Sparse Reconstruction"
% Ryota Tomioka and Masashi Sugiyama
% http://arxiv.org/abs/0904.0584
% 
% Copyright(c) 2009 Ryota Tomioka
% This software is distributed under the MIT license. See license.txt
 
 

function [xx, status]=dal(prob, xx0, A, AT, lambda, varargin)

opt=propertylist2struct(varargin{:});
opt=set_defaults(opt, 'aa', [],...
                      'tol', 1e-3, ...
                      'iter', 0, ...
                      'maxiter', 100,...
                      'eta', 1,...
                      'eps', 1e-4, ...
                      'eps_multp', 0.5,...
                      'eta_multp', 2, ...
                      'solver', 'nt', ...
                      'display',2);



if opt.display>0
  lstr=func2str(prob.floss.p); lstr=lstr(6:end-1);
  fprintf(['DAL ver0.97\n#samples=%d #variables=%d lambda=%g ' ...
           'loss=%s\n'],prob.mm,prob.nn,lambda, lstr);
end

floss = @(ww)prob.floss.p(A(ww(:)),prob.floss.args{:});
ffval = @(ww)floss(ww)+lambda*prob.freg(ww);

if opt.iter
  xx  = [xx0(:), ones(length(xx0(:)),opt.maxiter-1)*nan];
end

res  = nan*ones(size(1,opt.maxiter));
fval = nan*ones(size(1,opt.maxiter));

time0=cputime;
ww   = xx0;
gtmp = zeros(size(ww));
if isempty(opt.aa)
  aa = zeros(prob.mm,1);
else
  aa = opt.aa;
end

eta  = opt.eta;
epsl = sqrt(prob.mm)*opt.eps;
for ii=1:opt.maxiter-1
  %% Evaluate primal loss
  [fl, gg] = floss(ww);
  fval(ii) = fl+lambda*prob.freg(ww);

  %% Check stopping condition
  [ret,res(ii)]=prob.stopcond(prob.floss, ww, ...
                        A, AT, lambda, fl, gg, opt.tol);

  %% Display
  if opt.display>1
    fprintf('[[%d]] fval=%g res=%g\n', ii, fval(ii), res(ii));
  end

  if ret~=0
    break;
  end
  
  %% Solve minimization with respect to aa
  fun  = @(aa)prob.obj(aa, prob.floss, ww, A, AT, lambda, eta, opt.solver);
  switch(opt.solver)
   case {'nt','ntsv'}
    [aa,dfval,dgg,stat] = newton(fun, aa, epsl, opt.display>2);
   case 'cg'
    funh = @(xx,Hinfo)prob.hessMult(xx,A,eta,Hinfo);
    fh = {fun, funh};
    [aa,dfval,dgg,stat] = newton(fh, aa, epsl, opt.display>2);
   case 'qn'
    [aa,stat]=lbfgs(fun,aa,'epsg',epsl,'display',opt.display-1);
  end

  %% Update primal variable
  gtmp(:) = AT(aa);
  ww      = prob.softth(ww-eta*gtmp,eta*lambda);

  %% Update barrier parameter eta and tolerance parameter epsl
  eta     = eta*opt.eta_multp^(stat.ret==0);
  epsl    = epsl*opt.eps_multp^(stat.ret==0);
  if opt.iter
    xx(:,ii+1)=ww(:);
  end
 
end

res(ii+1:end)=[];
fval(ii+1:end)=[];
if opt.iter
  xx(:,ii+1:end)=[];
else
  xx = ww;
end


status=struct('aa', aa,...
              'niter',length(res),...
              'time', cputime-time0,...
              'res', res,...
              'opt', opt, ...
              'fval', fval);
